import argparse
import json
import pathlib
from typing import Any

from .api import evaluate, render, transpile


def parse_context(path: str | pathlib.Path | None, args: list[str]) -> dict[str, Any]:
    context = {}
    if path:
        path = pathlib.Path(path)
        context.update(json.loads(path.read_text()))
    for arg in args:
        if "=" not in arg:
            raise ValueError(f"invalid argument: {arg} (expected <key>=<value>)")
        key, value = arg.split("=", 1)
        try:
            context[key] = json.loads(value)
        except json.JSONDecodeError:
            context[key] = value
    return context


def cli() -> None:
    parser = argparse.ArgumentParser(description="Auryn metaprogramming engine")
    subparsers = parser.add_subparsers(dest="command", required=True)

    transpile_parser = subparsers.add_parser("transpile", help="transpile a template")
    transpile_parser.add_argument("path", help="template path")
    transpile_parser.add_argument("-l", "--load", default=None, help="additional meta-module path or name")
    transpile_parser.add_argument(
        "-n",
        "--no-common",
        action="store_true",
        default=False,
        help="do not load common meta-module",
    )
    transpile_parser.add_argument(
        "-s",
        "--standalone",
        action="store_true",
        default=False,
        help="generate standalone code",
    )

    render_parser = subparsers.add_parser("render", help="render a template")
    render_parser.add_argument("path", help="template path")
    render_parser.add_argument("-c", "--context", default=None, help="context path")
    render_parser.add_argument("-l", "--load", default=None, help="additional meta-module path or name")
    render_parser.add_argument(
        "-n",
        "--no-common",
        action="store_true",
        default=False,
        help="do not load common meta-module",
    )
    render_parser.add_argument(
        "context_kwargs",
        nargs="*",
        default=[],
        help="additional context as key=value pairs",
    )

    evaluate_parser = subparsers.add_parser("evaluate", help="evaluate a template")
    evaluate_parser.add_argument("path", help="template path")
    evaluate_parser.add_argument("-c", "--context", default=None, help="context path")
    evaluate_parser.add_argument(
        "context_kwargs",
        nargs="*",
        default=[],
        help="additional context as key=value pairs",
    )

    args = parser.parse_args()

    match args.command:
        case "transpile":
            code = transpile(
                pathlib.Path(args.path).absolute(),
                load=args.load,
                load_common=not args.no_common,
                standalone=args.standalone,
            )
            print(code)

        case "render":
            context = parse_context(args.context, args.context_kwargs)
            output = render(
                pathlib.Path(args.path).absolute(),
                context,
                load=args.load,
                load_common=not args.no_common,
            )
            print(output)

        case "evaluate":
            context = parse_context(args.context, args.context_kwargs)
            output = evaluate(
                pathlib.Path(args.path).absolute(),
                context,
            )
            print(output)



from __future__ import annotations

import pathlib
from typing import Iterator

from .utils import LINE_REGEX, trim


class Lines:

    def __init__(self, lines: list[Line]) -> None:
        self.lines = lines

    def __bool__(self) -> bool:
        return bool(self.lines)

    def __len__(self) -> int:
        return len(self.lines)

    def __iter__(self) -> Iterator[Line]:
        yield from self.lines

    @classmethod
    def from_file(cls, path: str | pathlib.Path) -> Lines:
        path = pathlib.Path(path)
        return cls.from_string(path.name, path.read_text())

    @classmethod
    def from_string(cls, name: str, string: str) -> Lines:
        lines: list[Line] = []
        stack: list[Line] = []
        for number, line_string in enumerate(trim(string), 1):
            if not (match := LINE_REGEX.match(line_string)):
                continue
            whitespace, content = match.groups()
            indent = len(whitespace)
            if not content:
                continue
            line = Line(name, number, indent, content)
            while stack and stack[-1].indent >= indent:
                stack.pop()
            if stack:
                stack[-1].children.lines.append(line)
            else:
                lines.append(line)
            stack.append(line)
        return cls(lines)

    def dedent(self, offset: int) -> Lines:
        for line in self.lines:
            line.dedent(offset)
        return self

    def align(self, to: int = 0) -> Lines:
        for line in self.lines:
            line.align(to)
        return self


class Line:

    def __init__(self, source: str, number: int, indent: int, content: str) -> None:
        self.source = source
        self.number = number
        self.indent = indent
        self.content = content
        self.children = Lines([])

    def __str__(self) -> str:
        return f"line {self.number} of {self.source}"

    def __repr__(self) -> str:
        return f"<{self}: {self.content}>"

    def dedent(self, offset: int) -> Line:
        self.indent = max(self.indent - offset, 0)
        self.children.dedent(offset)
        return self

    def align(self, to: int = 0) -> Line:
        self.dedent(self.indent - to)
        return self



from __future__ import annotations

import contextlib
import inspect
import pathlib
import re
import sys
from typing import Any, Callable, ClassVar, Iterator, Protocol

from .collect import collect_definitions, collect_global_references
from .interpolate import interpolate as interpolate_
from .lines import Line, Lines
from .utils import trim

type Transpiler = Callable[[Junk, str], None]
type MetaCallback = Callable[[Junk], None]


class MetaFunction(Protocol):
    def __call__(self, junk: Junk, *args: Any, **kwargs: Any) -> None: ...


META_REGEX = re.compile(
    r"""
    ^
    ([a-zA-Z_][a-zA-Z0-9_]*)
    (?:
        \(
        (.*?)
        \)
        |
        [ ]+
        (.*)
    )?
    $
    """,
    flags=re.VERBOSE,
)


class Junk:

    code_prefix: ClassVar[str] = "!"
    meta_prefix: ClassVar[str] = "%"
    builtins_directory: ClassVar[pathlib.Path] = pathlib.Path(__file__).parent / "builtins"
    load_common_by_default: ClassVar[bool] = True
    interpolate_by_default: ClassVar[bool] = True

    class StopEvaluation(Exception):
        pass

    def __init__(self, source: str, lines: Lines, path: str | pathlib.Path) -> None:
        self.source = source
        self.lines = lines
        self.path = pathlib.Path(path)
        self.code_indent = 0
        self.text_indent = 0
        self.code_lines: list[str] = []
        self.output: list[str] = []
        self.transpilers: dict[str, Transpiler] = {
            self.code_prefix: _transpile_code,
            self.meta_prefix: _transpile_meta,
            "": _transpile_text,
        }
        self.meta_namespace: dict[str, MetaFunction] = {
            "load": _load,
        }
        self.meta_state: dict[str, Any] = {}
        self.meta_callbacks: list[MetaCallback] = []
        self.eval_namespace: dict[str, Any] = {
            "emit": self.emit,
            "indent": self.indent,
            "StopEvaluation": self.StopEvaluation,
        }
        self.interpolation: str = "{ }"
        self.inline: bool = False
        self._active_lines: list[Line] = []

    def __repr__(self) -> str:
        if self.path:
            return f"<junk of {self.path} at {self.source}>"
        return f"<junk at {self.source}>"

    @classmethod
    def transpiler(cls, prefix: str = "") -> Callable[[Transpiler], Transpiler]:
        def decorator(transpiler: Transpiler) -> Transpiler:
            cls.transpilers[prefix] = transpiler
            return transpiler

        return decorator

    @classmethod
    def from_file(cls, path: str | pathlib.Path, *, stack_level: int = 0) -> Junk:
        source, source_path = cls._get_source(stack_level + 1)
        path = source_path.parent / path
        lines = Lines.from_file(path)
        return cls(source, lines, path)

    @classmethod
    def from_string(cls, string: str, *, stack_level: int = 0) -> Junk:
        source, path = cls._get_source(stack_level + 1)
        lines = Lines.from_string(source, string)
        return cls(source, lines, path)

    @classmethod
    def evaluate_standalone(
        cls,
        path: str | pathlib.Path,
        context: dict[str, Any] | None = None,
        /,
        *,
        stack_level: int = 0,
        **context_kwargs: Any,
    ) -> str:
        path = pathlib.Path(path)
        source, _ = cls._get_source(stack_level=stack_level + 1)
        junk = cls(source, Lines([]), path)
        junk.code_lines = path.read_text().splitlines()
        return junk.evaluate(context, junk=junk, **context_kwargs)

    @classmethod
    def _get_source(cls, stack_level: int) -> tuple[str, pathlib.Path]:
        frame = inspect.currentframe()
        for _ in range(stack_level + 1):
            frame = frame and frame.f_back
        if not frame:
            raise ValueError("unable to infer source")
        path = pathlib.Path(frame.f_code.co_filename)
        return f"{path.stem}:{frame.f_lineno}", path

    @property
    def line(self) -> Line:
        return self._active_lines[-1]

    def transpile(
        self,
        lines: Lines | None = None,
        load: str | pathlib.Path | dict[str, Any] | None = None,
        load_common: bool | None = None,
    ) -> str:
        if lines is None:
            lines = self.lines
            if load_common is None:
                load_common = self.load_common_by_default
            if load is not None:
                _load(self, load)
            if load_common:
                _load(self, "common")
        for line in lines:
            with self._set_active_line(line):
                for prefix, transpile in sorted(
                    self.transpilers.items(),
                    key=lambda x: len(x[0]),
                    reverse=True,
                ):
                    if line.content.startswith(prefix):
                        content = line.content.removeprefix(prefix).strip()
                        transpile(self, content)
                        break
                else:
                    raise ValueError(f"unable to transpile {line}")
        for callback in self.meta_callbacks:
            callback(self)
        return self.to_string()

    def to_string(self, standalone: bool = False) -> str:
        code = "\n".join(self.code_lines)
        if standalone:
            code = self._generate_intro(code)
        return code

    @contextlib.contextmanager
    def increase_code_indent(self) -> Iterator[None]:
        self.code_indent += 4
        try:
            yield
        finally:
            self.code_indent -= 4

    def emit_code(self, code: str) -> None:
        for code_line in trim(code):
            self.code_lines.append(self.code_indent * " " + code_line)

    def emit_text(
        self,
        indent: int,
        text: str,
        interpolate: bool | None = None,
        newline: bool = True,
    ) -> None:
        if interpolate is None:
            interpolate = self.interpolate_by_default
        if not interpolate:
            args = [repr(text)]
        else:
            args = []
            for snippet, is_code in interpolate_(text, self.interpolation):
                if is_code:
                    args.append(f"{snippet}")
                else:
                    args.append(repr(snippet))
        if not newline:
            args.append("newline=False")
        if self.inline:
            args.append("inline=True")
        self.emit_code(f'emit({indent}, {", ".join(args)})')

    def emit(
        self,
        indent: int,
        *args: Any,
        inline: bool = False,
        newline: bool = True,
    ) -> None:
        if inline:
            self.output.append("".join(map(str, args)))
        else:
            end = "\n" if newline else ""
            indent += self.text_indent
            self.output.append(f'{" " * indent}{"".join(map(str, args))}{end}')

    @contextlib.contextmanager
    def indent(self, indent: int) -> Iterator[None]:
        self.text_indent += indent
        try:
            yield
        finally:
            self.text_indent -= indent

    def evaluate(
        self,
        context: dict[str, Any] | None = None,
        /,
        **context_kwargs: Any,
    ) -> str:
        if context is not None:
            self.eval_namespace.update(context)
        self.eval_namespace.update(context_kwargs)
        with contextlib.suppress(self.StopEvaluation):
            exec(self.to_string(), self.eval_namespace)
        return "".join(self.output).rstrip()

    def interpolate(self, string: str) -> str:
        args = []
        for snippet, is_code in interpolate_(string, self.interpolation):
            if is_code:
                args.append(f"{snippet}")
            else:
                args.append(repr(snippet))
        if not args:
            return "''"
        if len(args) == 1:
            return f"str({args[0]})"
        return f'concat({", ".join(args)})'

    @contextlib.contextmanager
    def redirect_output(self) -> Iterator[list[str]]:
        output: list[str] = []
        prev_output, self.output = self.output, output
        try:
            yield output
        finally:
            self.output = prev_output

    @contextlib.contextmanager
    def _set_active_line(self, line: Line) -> Iterator[None]:
        self._active_lines.append(line)
        try:
            yield
        finally:
            self._active_lines.pop()

    def _generate_intro(self, code: str) -> str:
        paths: set[pathlib.Path] = set()
        for name in collect_global_references(code):
            if name not in self.eval_namespace:
                continue
            func = self.eval_namespace[name]
            if getattr(self, name, None) == func:
                continue
            while hasattr(func, "__wrapped__"):
                func = func.__wrapped__
            paths.add(pathlib.Path(func.__code__.co_filename))
        defs, imps = collect_definitions(code, paths)
        intro: list[str] = []
        for name, (what, whence) in imps.items():
            if whence is None:
                if name == what:
                    intro.append(f"import {name}")
                else:
                    intro.append(f"import {what} as {name}")
            elif name == what:
                intro.append(f"from {whence} import {name}")
            else:
                intro.append(f"from {whence} import {what} as {name}")
        if intro:
            intro.append("")
        for name, def_ in defs.items():
            intro.append(f"{def_}\n")
        return "\n".join(intro) + "\n" + code


def _transpile_code(junk: Junk, content: str) -> None:
    text_indent = junk.line.indent
    if text_indent:
        junk.emit_code(f"with indent({text_indent}):")
        junk.code_indent += 4
    junk.emit_code(content)
    with junk.increase_code_indent():
        junk.transpile(junk.line.children.dedent(junk.line.indent + 4))
    if text_indent > 0:
        junk.code_indent -= 4


def _transpile_meta(junk: Junk, content: str) -> None:
    match = META_REGEX.match(content)
    if not match:
        raise ValueError(
            f"expected meta function on {junk.line} to be '<function> [<argument>]' or '<function>(<arguments>)',"
            f"but got {content!r}"
        )
    name, args, arg = match.groups()
    meta_functions = [name for name, value in junk.meta_namespace.items() if callable(value)]
    if name not in meta_functions:
        raise ValueError(
            f"unknown meta function {name!r} on {junk.line} "
            f"(available meta functions are {', '.join(meta_functions)})"
        )
    if args:
        meta_code = f"{name}(junk, {args})"
    elif arg:
        meta_code = f"{name}(junk, {repr(arg)})"
    else:
        meta_code = f"{name}(junk)"
    eval(meta_code, {"junk": junk}, junk.meta_namespace)


def _transpile_text(junk: Junk, content: str) -> None:
    junk.emit_text(junk.line.indent, junk.line.content)
    junk.transpile(junk.line.children)


def _load(junk: Junk, target: str | pathlib.Path | dict[str, Any]) -> None:
    if isinstance(target, dict):
        namespace = target
    else:
        path = junk.path.parent / target
        if not path.exists():
            path = junk.builtins_directory / f"{target}.py"
            if not path.exists():
                raise ValueError(f"could not load {target!r} " "(neither file nor builtin module found)")
        sys_path = sys.path.copy()
        sys_path.append(str(path.parent))
        try:
            text = path.read_text()
            code = compile(text, str(path), "exec")
            namespace = {}
            exec(code, namespace)
        finally:
            sys.path = sys_path
    for key, value in namespace.items():
        if key.startswith("meta_"):
            name = key.removeprefix("meta_")
            junk.meta_namespace[name] = value
        if key.startswith("eval_"):
            name = key.removeprefix("eval_")
            junk.eval_namespace[name] = value.__get__(junk)
    if "on_load" in namespace:
        namespace["on_load"](junk)



import re

LINE_REGEX = re.compile(r"^(\s*)(.*)$")


def trim(string: str) -> list[str]:
    indent = None
    output = []
    for number, line in enumerate(string.rstrip().expandtabs().splitlines(), 1):
        if not line.strip():
            continue
        if indent is None:
            if not (match := LINE_REGEX.match(line)):
                continue
            whitespace, content = match.groups()
            indent = len(whitespace)
            output.append(content)
            continue
        prefix = line[:indent]
        if prefix and not prefix.isspace():
            raise ValueError(f"expected line {number} to start with {indent!r} spaces, " f"but got {prefix!r}")
        output.append(line[indent:])
    return output



from typing import Iterator


def interpolate(s: str, delimiters: str) -> Iterator[tuple[str, bool]]:
    a, b = delimiters.split(" ")
    if s == a or s == b or (a not in s and b not in s):
        yield s, False
        return
    L, aL, bL = len(s), len(a), len(b)
    i = 0
    text = []
    while i < L:
        if s[i : i + aL] == a:
            if s[i + aL : i + 2 * aL] == a:
                text.append(a)
                i += 2 * aL
            else:
                fr = i + aL
                to = skip_expression(s, a, b, fr)
                code = s[fr:to].strip()
                if text:
                    yield "".join(text), False
                    text.clear()
                yield code, True
                i = to + bL
        elif s[i : i + bL] == b:
            if s[i + bL : i + 2 * bL] == b:
                text.append(b)
                i += 2 * bL
            else:
                raise ValueError(f"unmatched {b!r} at offset {i}")
        else:
            text.append(s[i])
            i += 1
    if text:
        yield "".join(text), False


def skip_expression(s, a, b, i):
    L, aL, bL = len(s), len(a), len(b)
    depth = 1
    while i < L:
        if s[i : i + aL] == a:
            depth += 1
            i += aL
        elif s[i : i + bL] == b:
            depth -= 1
            if depth == 0:
                break
            i += bL
        elif s[i] in ["'", '"']:
            i = skip_string(s, i)
        else:
            i += 1
    else:
        raise ValueError(f"unmatched {a!r} at offset {i}")
    return i


def skip_string(s, i):
    L = len(s)
    q = s[i]
    i += 1
    while i < L:
        if s[i] == q:
            i += 1
            break
        if s[i] == "\\":
            i += 2
        else:
            i += 1
    else:
        raise ValueError(f"unterminated quote at offset {i}")
    return i



import contextlib
import pathlib
from typing import Any

from auryn import Junk, Lines

UNDEFINED = object()


def meta_include(junk: Junk, path: str | pathlib.Path) -> None:
    included_junk = junk.from_file(junk.path.parent / path)
    with contextlib.suppress(IndexError):
        included_junk.lines.align(junk.line.indent)
    included_junk.meta_state = junk.meta_state
    included_junk.transpile()
    junk.emit_code(included_junk.to_string())


def meta_define(junk: Junk, name: str) -> None:
    definitions = junk.meta_state.setdefault("definitions", {})
    definitions[name] = junk.line.children


def meta_insert(junk: Junk, name: str) -> None:
    definitions = junk.meta_state.get("definitions", {})
    junk.transpile(definitions[name].align(junk.line.indent))


def meta_extend(junk: Junk, template: str | pathlib.Path) -> None:
    def replace_code(junk: Junk) -> None:
        junk.code_lines.clear()
        junk.code_indent = 0
        meta_include(junk, template)

    junk.meta_callbacks.append(replace_code)


def meta_interpolate(junk: Junk, delimiters: str) -> None:
    if junk.line.children:
        prev_delimiters, junk.interpolation = junk.interpolation, delimiters
        junk.transpile(junk.line.children.align(junk.line.indent))
        junk.interpolation = prev_delimiters
    else:
        junk.interpolation = delimiters


def meta_raw(junk: Junk) -> None:
    if junk.line.children:
        junk.emit_text(0, to_string(junk.line.children.align(junk.line.indent)))
    else:
        junk.transpilers = {"": emit_raw}


def to_string(lines: Lines) -> str:
    text = []
    for line in lines:
        text.append(" " * line.indent + line.content)
        text.append(to_string(line.children))
    return "\n".join(text)


def emit_raw(junk: Junk, content: str) -> None:
    junk.emit_code(f"emit({junk.line.indent}, {content!r})")
    junk.transpile(junk.line.children)


def meta_stop(junk: Junk) -> None:
    junk.emit_code(f"raise {junk.StopEvaluation.__name__}()")


def meta_param(junk: Junk, name: str, default: Any = UNDEFINED) -> None:
    parameters = junk.meta_state.setdefault("parameters", {})
    parameters[name] = default if default is not UNDEFINED else "<required>"
    if default is UNDEFINED:
        message = f"missing required parameter {name!r} if {junk.source}"
        junk.emit_code(
            f"""
            if {name!r} not in globals():
                raise ValueError({message!r})
    I       """
        )
    else:
        junk.emit_code(
            f"""
            try:
                {name}
            except NameError:
                {name} = {default!r}
            """
        )


def meta_inline(junk: Junk) -> None:
    prev_inline = junk.inline
    try:
        junk.emit_text(junk.line.indent, "", newline=False)
        junk.inline = True
        junk.transpile(junk.line.children)
        junk.inline = False
        junk.emit_text(0, "")
    finally:
        junk.inline = prev_inline


def meta_strip(junk: Junk, suffix: str) -> None:
    junk.emit_code(f"strip({suffix!r})")


def eval_strip(junk: Junk, suffix: str) -> None:
    junk.output[-1] = junk.output[-1].rstrip().strip(suffix)


def eval_camel_case(junk: Junk, name: str) -> str:
    return "".join(word.capitalize() for word in name.split("_"))


def eval_concat(*args: Any) -> str:
    return "".join(map(str, args))



import contextlib
import os
import pathlib
import subprocess
from typing import Iterator

from auryn import Junk


def on_load(junk: Junk) -> None:
    junk.emit_code(
        """
        try:
            __import__("os").chdir(root)
        except NameError:
            __import__("os").chdir(junk.path.parent)
        """
    )


def meta_d(
    junk: Junk,
    name: str,
    source: str | pathlib.Path | None = None,
    render: bool = False,
    interpolate: bool | None = None,
) -> None:
    junk.emit_code(f"with directory({junk.interpolate(name)}):")
    with junk.increase_code_indent():
        if source:
            root = junk.path.parent / source
            print(root)
            for path in root.rglob("*"):
                if path.is_file():
                    name = str(path.relative_to(root))
                    meta_f(junk, name, path, render, interpolate)
        if junk.line.children:
            junk.transpile(junk.line.children.align(junk.line.indent))
        else:
            junk.emit_code("pass")


@contextlib.contextmanager
def eval_directory(junk: Junk, name: str) -> Iterator[None]:
    cwd = os.getcwd()
    path = pathlib.Path(name)
    path.mkdir(parents=True, exist_ok=True)
    os.chdir(path)
    yield
    os.chdir(cwd)


def meta_f(
    junk: Junk,
    name: str,
    source: str | pathlib.Path | None = None,
    render: bool = False,
    interpolate: bool | None = None,
) -> None:
    junk.emit_code(f"with file({junk.interpolate(name)}):")
    with junk.increase_code_indent():
        if source:
            if render:
                source_junk = junk.from_file(junk.path.parent / source)
                source_junk.meta_state = junk.meta_state
                source_junk.transpile()
                junk.emit_code(source_junk.to_string())
            else:
                source_text = (junk.path.parent / source).read_text()
                junk.emit_text(0, source_text, interpolate=interpolate)
        elif junk.line.children:
            junk.transpile(junk.line.children.align(junk.line.indent))
        else:
            junk.emit_code("pass")


@contextlib.contextmanager
def eval_file(junk: Junk, name: str) -> Iterator[None]:
    path = pathlib.Path(name)
    path.parent.mkdir(parents=True, exist_ok=True)
    with junk.redirect_output() as output:
        yield
        path.write_text("".join(output).strip())


def meta_x(
    junk,
    command,
    *,
    into: str | None = None,
    stderr_into: str | None = None,
    status_into: str | None = None,
    timeout: int | None = None,
    strict: bool | None = None,
):
    args = [junk.interpolate(command)]
    if timeout:
        args.append(f"timeout={timeout!r}")
    if strict:
        args.append("strict=True")
    retvals = [
        into if into else "_",
        stderr_into if stderr_into else "_",
        status_into if status_into else "_",
    ]
    junk.emit_code(f"{', '.join(retvals)} = shell({', '.join(args)})")


def eval_shell(
    junk: Junk,
    command: str,
    *,
    timeout: int | None = None,
    strict: bool | None = None,
) -> tuple[str, str, int]:
    result = subprocess.run(command, shell=True, capture_output=True, timeout=timeout)
    if strict and result.returncode:
        raise RuntimeError(f"failed to run {command!r}: " f"[{result.returncode}] {result.stderr.decode()}")
    return result.stdout.decode(), result.stderr.decode(), result.returncode



import ast
import builtins
import pathlib
from types import CodeType

BUILTIN_NAMES = set(vars(builtins))


class DefinitionCollector(ast.NodeTransformer):

    def __init__(self) -> None:
        self.evals: dict[str, str] = {}
        self.defs: dict[str, str] = {}
        self.imps: dict[str, tuple[str, str | None]] = {}

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        if node.name.startswith("eval_"):
            node.name = node.name.removeprefix("eval_")
            node.args.args = node.args.args[1:]
            self.evals[node.name] = ast.unparse(node)
        else:
            self.defs[node.name] = ast.unparse(node)
        return node

    def visit_Import(self, node: ast.Import) -> ast.Import:
        for alias in node.names:
            self.imps[alias.asname or alias.name] = alias.name, None
        return node

    def visit_ImportFrom(self, node: ast.ImportFrom) -> ast.ImportFrom:
        for alias in node.names:
            self.imps[alias.asname or alias.name] = alias.name, node.module
        return node


def collect_definitions(
    code: str, paths: set[pathlib.Path]
) -> tuple[dict[str, str], dict[str, tuple[str, str | None]]]:
    evals: dict[str, str] = {}
    defs: dict[str, str] = {}
    imps: dict[str, tuple[str, str | None]] = {}
    for path in paths:
        dc = DefinitionCollector()
        dc.visit(ast.parse(path.read_text()))
        evals |= dc.evals
        defs |= dc.defs
        imps |= dc.imps
    used_defs: dict[str, str] = {}
    used_imps: dict[str, tuple[str, str | None]] = {}
    for name in collect_global_references(code):
        if name not in evals:
            continue
        code = used_defs[name] = evals.pop(name)
        collect_deps(code, defs, imps, used_defs, used_imps)
    return used_defs, used_imps


def collect_deps(
    code: str,
    defs: dict[str, str],
    imps: dict[str, tuple[str, str | None]],
    used_defs: dict[str, str],
    used_imps: dict[str, tuple[str, str | None]],
) -> None:
    for name in collect_global_references(code):
        if name in imps:
            used_imps[name] = imps.pop(name)
        elif name in defs:
            code = used_defs[name] = defs.pop(name)
            collect_deps(code, defs, imps, used_defs, used_imps)


def collect_global_references(code: str | CodeType) -> set[str]:
    if isinstance(code, str):
        code = compile(code, "", "exec")
    names = set(code.co_names)
    for const in code.co_consts:
        if isinstance(const, CodeType):
            names |= collect_global_references(const)
    return names - BUILTIN_NAMES



import pathlib
from typing import Any

from .junk import Junk


def transpile(
    template: str | pathlib.Path,
    /,
    *,
    standalone: bool = False,
    load: str | pathlib.Path | dict[str, Any] | None = None,
    load_common: bool | None = None,
) -> str:
    if isinstance(template, str) and "\n" in template:
        junk = Junk.from_string(template, stack_level=1)
    else:
        junk = Junk.from_file(template, stack_level=1)
    junk.transpile(load=load, load_common=load_common)
    return junk.to_string(standalone=standalone)


def render(
    template: str | pathlib.Path,
    context: dict[str, Any] | None = None,
    /,
    *,
    load: str | pathlib.Path | dict[str, Any] | None = None,
    load_common: bool | None = None,
    **context_kwargs: Any,
) -> str:
    if isinstance(template, str) and "\n" in template:
        junk = Junk.from_string(template, stack_level=1)
    else:
        junk = Junk.from_file(template, stack_level=1)
    junk.transpile(load=load, load_common=load_common)
    return junk.evaluate(context, **context_kwargs)


def evaluate(
    path: str | pathlib.Path,
    context: dict[str, Any] | None = None,
    /,
    **context_kwargs: Any,
) -> str:
    return Junk.evaluate_standalone(path, context, stack_level=1, **context_kwargs)
